/*************************************************************************\
* Copyright (c) 2006 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* This software is distributed subject to a Software License Agreement
* found in file LICENSE that is included with this distribution.
\*************************************************************************/

/*
 * Flash memory programmer
 * Allows flash updates using standard channel access operations.
 */

#include <string.h>
#include <epicsStdioRedirect.h>
#include <epicsStdlib.h>
#include <epicsString.h>
#include <epicsThread.h>
#include <asynDriver.h>
#include <asynOctet.h>
#include "devFlashBurner.h"
#include "flashBurner.h"

/*
 * Per-driver information
 */
struct dpvt {
    asynInterface                 common;
    asynInterface                 octet;
    struct drvFlashBurnerMethods  methods;
    void                         *drvpvt;
};

/*
 * asynCommon methods
 */
static asynStatus
connect(void *drvPvt, asynUser *pasynUser)
{
    pasynManager->exceptionConnect(pasynUser);
    return asynSuccess;
}

static asynStatus
disconnect(void *drvPvt, asynUser *pasynUser)
{
    pasynManager->exceptionDisconnect(pasynUser);
    return asynSuccess;
}

static void
report(void *drvPvt, FILE *fp, int level)
{
    struct dpvt *dpvt = (struct dpvt *)drvPvt;

    if (dpvt->methods.report)
        (*dpvt->methods.report)(dpvt->drvpvt, fp, level);
}

static struct asynCommon asynCommonMethods = { report, connect, disconnect };

/*
 * asynOctet methods
 */
static asynStatus
asynOctetWrite(void *drvPvt, asynUser *pasynUser,
                    const char *data, size_t numchars, size_t *nbytesTransfered)
{
    struct dpvt *dpvt = (struct dpvt *)drvPvt;
    struct flashBurnerWaveformData *pwd = (struct flashBurnerWaveformData *)data;
    asynStatus status;
    int byteCount;
    unsigned int offset;
    unsigned int checksum, hChecksum;
    int i;
    if (numchars != sizeof(*pwd)) {   
        epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
                            "Bad write request (numchars=%u, expected %u)",
                                                    (unsigned int)numchars,
                                                    (unsigned int)sizeof(*pwd));
            
        return asynError;
    }
    if (sscanf(pwd->header, FLASH_HEADER_FORMAT, &byteCount, &offset, &hChecksum) != 3) {
        epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
                                                                "Bad header");
        return asynError;
    }
    if ((byteCount < 0) || (byteCount > FLASH_PAGE_SIZE)) {
        epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
                            "Bad write request (byteCount=%d)", byteCount);
        return asynError;
    }
    checksum = 0;
    for (i = 0 ; i < FLASH_PAGE_SIZE ; i++)
        checksum += (unsigned char)pwd->flashData[i];
    if (checksum != hChecksum) {
        epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
                                    "Bad write request (checksum mismatch)");
        return asynError;
    }
    status = dpvt->methods.program(dpvt->drvpvt, pasynUser, offset,
                                                            pwd->flashData,
                                                            byteCount);
    if (status == asynSuccess)
        *nbytesTransfered = numchars;
    return status;
}

static asynOctet asynOctetMethods = { asynOctetWrite };

/*
 * Register a flash burner driver
 */
int
devFlashBurnerRegister(const char *name,
                       const struct drvFlashBurnerMethods *methods,
                       void *drvpvt)
{
    struct dpvt *dpvt;

    if (methods->program == NULL) {
        printf("Driver did not provide a program method!\n");
        return -1;
    }
    dpvt = calloc(1, sizeof *dpvt);
    if (dpvt == NULL) {
        printf("Can't allocate memory for flash burner %s\n", name);
        return -1;
    }
    dpvt->methods = *methods;
    dpvt->drvpvt = drvpvt;

    /*
     * Register with ASYN
     */
    dpvt->common.interfaceType = asynCommonType;
    dpvt->common.pinterface  = &asynCommonMethods;
    dpvt->common.drvPvt = dpvt;
    dpvt->octet.interfaceType = asynOctetType;
    dpvt->octet.pinterface  = &asynOctetMethods;
    dpvt->octet.drvPvt = dpvt;
    if ((pasynManager->registerPort(name,
               ASYN_CANBLOCK,
               1,
               epicsThreadPriorityLow,
               epicsThreadGetStackSize(epicsThreadStackSmall)) != asynSuccess)
     || (pasynManager->registerInterface(name, &dpvt->common) != asynSuccess)
     || (pasynOctetBase->initialize(name, &dpvt->octet, 0, 0, 0) != asynSuccess)) {

        printf("Can't register with ASYN\n");
        return -1;
    }
    return 0;
}
