/*************************************************************************\
* Copyright (c) 2006 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* This software is distributed subject to a Software License Agreement
* found in file LICENSE that is included with this distribution.
\*************************************************************************/

/*
 * ASYN Int32 driver for ColdFire General Purpose Timer data ports
 */

#include <bsp.h>
#include <epicsStdio.h>
#include <epicsExport.h>
#include <cantProceed.h>

#include <asynDriver.h>
#include <asynInt32.h>

#define INPUT_DATA    MCF5282_GPTA_GPTPORT
#define INPUT_DDR     MCF5282_GPTA_GPTDDR
#define OUTPUT_DATA   MCF5282_GPTB_GPTPORT
#define OUTPUT_DDR    MCF5282_GPTB_GPTDDR

/*
 * Driver private storage
 */
typedef struct drvPvt {
    asynInterface common;
    asynInterface int32;
} drvPvt;

/*
 * asynCommon methods
 */
static void
report(void *pvt, FILE *fp, int details)
{
    if (details >= 1)
        printf("  Input port(%p):%#x   Output port(%p):%#x\n",
                            &INPUT_DATA, INPUT_DATA, &OUTPUT_DATA, OUTPUT_DATA);
}

static asynStatus
connect(void *pvt, asynUser *pasynUser)
{
    pasynManager->exceptionConnect(pasynUser);
    return asynSuccess;
}

static asynStatus
disconnect(void *pvt, asynUser *pasynUser)
{
    pasynManager->exceptionDisconnect(pasynUser);
    return asynSuccess;
}
static asynCommon commonMethods = { report, connect, disconnect };

/*
 * asynInt32 methods
 */
static asynStatus
int32Write(void *pvt, asynUser *pasynUser, epicsInt32 value)
{
    asynStatus status;
    int address;

    if ((status = pasynManager->getAddr(pasynUser, &address)) != asynSuccess)
        return status;
    switch (address) {
    case 1: OUTPUT_DATA = value;        break;
    default:
        epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
                                        "Invalid write address %d", address);
        return asynError;
    }
    return asynSuccess;
}

static asynStatus
int32Read(void *pvt, asynUser *pasynUser, epicsInt32 *value)
{
    asynStatus status;
    int address;

    if ((status = pasynManager->getAddr(pasynUser, &address)) != asynSuccess)
        return status;
    switch (address) {
    case 0: *value = INPUT_DATA;         break;
    case 1: *value = OUTPUT_DATA;        break;
    default:
        epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
                                        "Invalid read address %d", address);
        return asynError;
    }
    return asynSuccess;
}
static asynInt32 int32Methods = { int32Write, int32Read };

/*
 * Register ourself with ASYN
 */
static void devGptPortsRegistrar(void)
{
    drvPvt *pdrvPvt;
    asynStatus status;
    const char *portName = "gptPorts";

    pdrvPvt = callocMustSucceed(sizeof(drvPvt), 1, "gptPorts");

    status = pasynManager->registerPort(portName, ASYN_MULTIDEVICE, 1, 0, 0);
    if(status != asynSuccess) {
        printf("registerDriver failed\n");
        return;
    }

    pdrvPvt->common.interfaceType = asynCommonType;
    pdrvPvt->common.pinterface  = &commonMethods;
    pdrvPvt->common.drvPvt = pdrvPvt;
    status = pasynManager->registerInterface(portName, &pdrvPvt->common);
    if (status != asynSuccess) {
        printf("registerInterface failed\n");
        return;
    }

    pdrvPvt->int32.interfaceType = asynInt32Type;
    pdrvPvt->int32.pinterface  = &int32Methods;
    pdrvPvt->int32.drvPvt = pdrvPvt;
    status = pasynInt32Base->initialize(portName, &pdrvPvt->int32);
    if (status != asynSuccess) {
        printf("pasynInt32Base initialize failed\n");
        return;
    }

    INPUT_DDR &= ~0xF;
    OUTPUT_DDR |= 0xF;
}
epicsExportRegistrar(devGptPortsRegistrar);
