/*
 * Altera FPGA configuration devices (EPCSxxx)
 */
#include <epicsStdioRedirect.h>
#include <epicsStdlib.h>
#include <epicsString.h>
#include <epicsThread.h>
#include <epicsInterrupt.h>
#include <epicsExport.h>
#include <cantProceed.h>
#include <devFlashBurner.h>
#include <ctype.h>
#include <bsp.h>

/*
 * I/O port assignments
 */
#define DATAREG_ASDI     MCF5282_GPTA_GPTPORT
#define DDRREG_ASDI      MCF5282_GPTA_GPTDDR
#define MASK_ASDI        (1<<3)
#define DATAREG_nCS      MCF5282_GPTA_GPTPORT
#define DDRREG_nCS       MCF5282_GPTA_GPTDDR
#define MASK_nCS         (1<<2)
#define DATAREG_DCLK     MCF5282_GPTA_GPTPORT
#define DDRREG_DCLK      MCF5282_GPTA_GPTDDR
#define MASK_DCLK        (1<<1)
#define DATAREG_DATA     MCF5282_GPTA_GPTPORT
#define DDRREG_DATA      MCF5282_GPTA_GPTDDR
#define MASK_DATA        (1<<0)
#define DATAREG_NCONFIG     MCF5282_GPTB_GPTPORT
#define DDRREG_NCONFIG      MCF5282_GPTB_GPTDDR
#define MASK_NCONFIG        (1<<1)
#define DATAREG_CONF_DONE   MCF5282_GPTB_GPTPORT
#define DDRREG_CONF_DONE    MCF5282_GPTB_GPTDDR
#define MASK_CONF_DONE      (1<<3)

/*
 * EPCS identification codes
 */
#define IDCODE_EPCS1          0x10
#define IDCODE_EPCS4          0x12
#define IDCODE_EPCS16         0x14
#define IDCODE_EPCS64         0x16

/*
 * EPCS parameters
 */
#define EPCS_PAGESIZE          256

/*
 * EPCS operation codes
 */
#define OPCODE_WRITE_ENABLE    0x06
#define OPCODE_WRITE_DISABLE   0x04
#define OPCODE_READ_STATUS     0x05
#define OPCODE_READ_BYTES      0x03
#define OPCODE_READ_SILICON_ID 0xAB
#define OPCODE_WRITE_STATUS    0x01
#define OPCODE_WRITE_BYTES     0x02
#define OPCODE_ERASE_BULK      0xC7
#define OPCODE_ERASE_SECTOR    0xD8

struct flashInfo {
    unsigned int    capacity;
    unsigned int    eraseHighWater;
    unsigned int    sectorSize;
};

static void
initializeHardware(void)
{
    int key = epicsInterruptLock();
    DATAREG_nCS  |=  MASK_nCS;
    DATAREG_DCLK &= ~MASK_DCLK;
    DDRREG_ASDI  |=  MASK_ASDI;
    DDRREG_nCS   |=  MASK_nCS;
    DDRREG_DCLK  |=  MASK_DCLK;
    DDRREG_DATA  &= ~MASK_DATA;
    epicsInterruptUnlock(key);
}

static void
shiftBytes(int nbytes, const unsigned char *obuf, unsigned char *ibuf)
{
    int b;
    int s;
    int key;

    key = epicsInterruptLock();
    DATAREG_nCS &= ~MASK_nCS;
    epicsInterruptUnlock(key);
    while (nbytes--) {
        s = *obuf++;
        for (b = 0 ; b < 8 ; b++) {
            key = epicsInterruptLock();
            if (s & 0x80)
                DATAREG_ASDI |= MASK_ASDI;
            else
                DATAREG_ASDI &= ~MASK_ASDI;
            s <<= 1;
            if (DATAREG_DATA & MASK_DATA)
                s |= 1;
            DATAREG_DCLK |= MASK_DCLK;
            asm volatile ("nop");
            DATAREG_DCLK &= ~MASK_DCLK;
            epicsInterruptUnlock(key);
        }
        if (ibuf)
            *ibuf++ = s;
    }
    key = epicsInterruptLock();
    DATAREG_nCS |= MASK_nCS;
    epicsInterruptUnlock(key);
}

static void
csDelay(void)
{
    int i;

    for (i = 0 ; i < 10 ; i++)
        asm volatile ("nop");
}

static int
readSiliconID(void)
{
    static const unsigned char cmd[] = { OPCODE_READ_SILICON_ID, 0, 0, 0, 0 };
    unsigned char reply[sizeof cmd];

    shiftBytes(sizeof cmd, cmd, reply);
    return reply[(sizeof cmd) - 1];
}

static int
readStatus(void)
{
    static const unsigned char cmd[] = { OPCODE_READ_STATUS, 0 };
    unsigned char reply[sizeof cmd];

    shiftBytes(sizeof cmd, cmd, reply);
    return reply[(sizeof cmd) - 1];
}

static int
writeEnable(void)
{
    static const unsigned char cmd[] = { OPCODE_WRITE_ENABLE };
    int s;

    shiftBytes(sizeof cmd, cmd, NULL);
    csDelay();
    if (((s = readStatus()) & 0x2) == 0) {
        printf("EPCS write enable failed (status = %2.2X)\n", s);
        return asynError;
    }
    return asynSuccess;
}

static int
eraseSector(struct flashInfo *f, unsigned int address)
{
    unsigned char cmd[4];
    int pass;
    
    if (address != f->eraseHighWater) {
        printf("Warning, erase %#x, but erase high water %#x\n",
                                                         address,
                                                         f->eraseHighWater);
    }
    if (writeEnable() != asynSuccess)
        return asynError;
    cmd[0] = OPCODE_ERASE_SECTOR;
    cmd[1] = address >> 16;
    cmd[2] = address >> 8;
    cmd[3] = address;
    shiftBytes(sizeof cmd, cmd, NULL);
    epicsThreadSleep(0.5);
    for (pass = 0 ; pass < 20 ; pass++) {
        if ((readStatus() & 0x1) == 0) {
            f->eraseHighWater = address + f->sectorSize - (address % f->sectorSize);
            return asynSuccess;
        }
        epicsThreadSleep(0.1);
    }
    printf("EPCS erase did not complete!\n");
    return asynError;
}

/*
 * Display contents of 16 bytes at the specified address
 */
static void
showContents(int address)
{
    int i;
    unsigned char iobuf[16+4];

    iobuf[0] = OPCODE_READ_BYTES;
    iobuf[1] = address >> 16;
    iobuf[2] = address >> 8;
    iobuf[3] = address;
    shiftBytes(sizeof iobuf, iobuf, iobuf);

    printf("%6.6X", address);
    for (i = 0 ; i < 16 ; i++)
        printf(" %2.2X", iobuf[4+i]);
    printf("  ");
    for (i = 0 ; i < 16 ; i++)
        printf("%c", isprint(iobuf[4+i]) ? iobuf[4+i] : '.');
    printf("\n");
}

static void
epcsReport(void *dpvt, FILE *fp, int level)
{
    struct flashInfo *f = dpvt;

    fprintf(fp, "   %d bytes\n", f->capacity);
}

static asynStatus
epcsProgram(void *dpvt, asynUser *pasynUser, unsigned int address, void *buf, int count)
{
    struct flashInfo *f = dpvt;
    unsigned char *cp = buf;
    unsigned int startAddress;
    int i, j;
    unsigned char cmd[EPCS_PAGESIZE+4];
    unsigned char chk[EPCS_PAGESIZE+4];

    if (count == 0) {
        f->eraseHighWater = 0;
        return asynSuccess;
    }
    i = 0;
    startAddress = address;
    for (;;) {
        if (address >= f->eraseHighWater) {
            if (eraseSector(f, address) != asynSuccess)
                return asynError;
        }
        if (address >= f->eraseHighWater) {
            epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
                    "Flash Burner: Erase didn't clear the right location.\n");
            return asynError;
        }
        if (i == 0) {
            cmd[0] = OPCODE_WRITE_BYTES;
            cmd[1] = address >> 16;
            cmd[2] = address >> 8;
            cmd[3] = address;
            startAddress = address;
        }
        cmd[i++ + 4] = *cp++;
        address++;
        count--;
        if ((count == 0) || ((address % EPCS_PAGESIZE) == 0)) {
            int pass = 0;
            if (writeEnable() != asynSuccess)
                return asynError;
            shiftBytes(4 + i, cmd, NULL);
            csDelay();
            for (;;) {
                epicsThreadSleep(0.002);
                if ((readStatus() & 0x1) == 0)
                    break;
                if (++pass == 50) {
                    epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
                            "Flash Burner: Programming did not complete.\n");
                    return asynError;
                }
            }
            cmd[0] = OPCODE_READ_BYTES;
            shiftBytes(4 + i, cmd, chk);
            for (j = 0 ; j < i ; j++) {
                if (chk[4+j] != cmd[4+j]) {
                    printf("Wrote %2.2X to %6.6X, but read %2.2X\n",
                                            cmd[j], startAddress + j, chk[j]);
                    return asynError;
                }
            }
            if (count == 0)
                return asynSuccess;
            i = 0;
        }
    }
}

static struct drvFlashBurnerMethods methods = {
   epcsReport,
   epcsProgram
};

static int
configure(const char *name, unsigned int displayLimit)
{
    int siliconID;
    struct flashInfo *f;
    unsigned int displayAddress;

    f = callocMustSucceed(1, sizeof (*f), "EPCS configure");
    initializeHardware();
    siliconID = readSiliconID();
    switch (siliconID) {
    default:
        printf("EPCS flash driver doesn't know about identification code %#x\n", siliconID);
        free(f);
        return asynError;
    case IDCODE_EPCS1:  f->capacity =  128*1024; f->sectorSize = 32768; break;
    case IDCODE_EPCS4:  f->capacity =  512*1024; f->sectorSize = 65536; break;
    case IDCODE_EPCS16: f->capacity = 2048*1024; f->sectorSize = 65536; break;
    case IDCODE_EPCS64: f->capacity = 8192*1024; f->sectorSize = 65536; break;
    }
    if (devFlashBurnerRegister(name, &methods, f) < 0) {
        free(f);
        return asynError;
    }
    printf("FPGA configuration device is EPCS%d (%d bytes per sector)\n",
                                        f->capacity/(128*1024), f->sectorSize);
        displayLimit = (displayLimit + 15) / 16;
    for (displayAddress = 0 ; displayAddress < displayLimit ; displayAddress++)
        showContents(displayAddress * 16);
    return asynSuccess;
}

/*
 * IOC shell command registration
 */
#include <iocsh.h>
static const iocshArg drvEPCSFlashBurnerConfigureArg0 = { "name",iocshArgString};
static const iocshArg drvEPCSFlashBurnerConfigureArg1 = { "display limit",iocshArgInt};
static const iocshArg *drvEPCSFlashBurnerConfigureArgs[] = {
           &drvEPCSFlashBurnerConfigureArg0, &drvEPCSFlashBurnerConfigureArg1 };
static const iocshFuncDef drvEPCSFlashBurnerConfigureFuncDef = {
              "drvEPCSFlashBurnerConfigure",2,drvEPCSFlashBurnerConfigureArgs
};
static void drvEPCSFlashBurnerConfigureCallFunc(const iocshArgBuf *args)
{
    configure(args[0].sval, args[1].ival);
}

/*
 * Force FPGA to reload configuration
 */
static const iocshFuncDef forceFPGAresetFuncDef = {"forceFPGAreset",0,NULL};
static void forceFPGAresetCallFunc(const iocshArgBuf *args)
{
    volatile int i;

    printf("========== RELOAD FPGA ========== RESET IOC ===========\n");
    printf("DATAREG_CONF_DONE:%x  DDRREG_CONF_DONE:%x\n", DATAREG_CONF_DONE, DDRREG_CONF_DONE);
    epicsThreadSleep(1.0);
    epicsInterruptLock();
    DDRREG_ASDI  &= ~MASK_ASDI;
    DDRREG_nCS   &= ~MASK_nCS;
    DDRREG_DCLK  &= ~MASK_DCLK;
    DATAREG_NCONFIG &= ~MASK_NCONFIG;
    DDRREG_NCONFIG |= MASK_NCONFIG;
    for (i = 0 ; i < 1000 ; i++)
        continue;
    DDRREG_NCONFIG &= ~MASK_NCONFIG;
    for (i = 0 ; (DATAREG_CONF_DONE & MASK_CONF_DONE) == 0 ; i++) {
        if (i >= 10000000) {
            printk("!!!!!!! FPGA CONF_DONE (%x %x) DID NOT GO HIGH!!!!!!!!\n\n\n", DATAREG_CONF_DONE, DDRREG_CONF_DONE);
            break;
        }
    }
    bsp_reset(0);
}
static void
drvEPCSFlashBurnerRegisterCommands(void)
{
    iocshRegister(&drvEPCSFlashBurnerConfigureFuncDef,drvEPCSFlashBurnerConfigureCallFunc);
    iocshRegister(&forceFPGAresetFuncDef,forceFPGAresetCallFunc);
}
epicsExportRegistrar(drvEPCSFlashBurnerRegisterCommands);
